import json
import tempfile
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

import pandas as pd
import pytest
from opensearchpy import Document, Keyword, OpenSearch, Text

from test.integration.connectors.utils.constants import DESTINATION_TAG, SOURCE_TAG
from test.integration.connectors.utils.docker import HealthCheck, container_context
from test.integration.connectors.utils.validation import (
    ValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.error import (
    DestinationConnectionError,
    SourceConnectionError,
)
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.elasticsearch.opensearch import (
    CONNECTOR_TYPE,
    OpenSearchAccessConfig,
    OpenSearchConnectionConfig,
    OpenSearchDownloader,
    OpenSearchDownloaderConfig,
    OpenSearchIndexer,
    OpenSearchIndexerConfig,
    OpenSearchUploader,
    OpenSearchUploaderConfig,
    OpenSearchUploadStager,
    OpenSearchUploadStagerConfig,
)

SOURCE_INDEX_NAME = "movies"
DESTINATION_INDEX_NAME = "elements"


class Movie(Document):
    title = Text(fields={"raw": Keyword()})
    year = Text()
    director = Text()
    cast = Text()
    genre = Text()
    wiki_page = Text()
    ethnicity = Text()
    plot = Text()

    class Index:
        name = SOURCE_INDEX_NAME

    def save(self, **kwargs):
        return super(Movie, self).save(**kwargs)


@contextmanager
def get_client() -> Generator[OpenSearch, None, None]:
    with OpenSearch(
        hosts=[{"host": "localhost", "port": 9200}],
        http_auth=("admin", "admin"),
        use_ssl=True,
        verify_certs=False,
        ssl_show_warn=False,
    ) as client:
        yield client


def get_index_count(client: OpenSearch, index_name: str) -> int:
    count_resp = client.cat.count(index=index_name, params={"format": "json"})
    return int(count_resp[0]["count"])


def wait_for_write(
    client: OpenSearch, index_name: str, expected_count: int, timeout: int = 30, interval: int = 1
) -> None:
    current_count = get_index_count(client, index_name)
    start = time.time()
    while time.time() - start < timeout:
        print(f"waiting for current count ({current_count}) to match expected {expected_count}")
        time.sleep(interval)
        current_count = get_index_count(client, index_name)
        if current_count == expected_count:
            return
    raise TimeoutError("Timed out while waiting for write to sync")


def validate_count(
    client: OpenSearch, index_name: str, expected_count: int, retries: int = 10, interval: int = 1
) -> None:
    current_count = get_index_count(client, index_name)
    if current_count == expected_count:
        return
    tries = 0
    while tries < retries:
        print(
            f"retrying validation to check if expected count "
            f"{expected_count} will match current count {current_count}"
        )
        time.sleep(interval)
        current_count = get_index_count(client, index_name)
        if current_count == expected_count:
            break
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from index: {current_count}"
    )


@pytest.fixture
def source_index(movies_dataframe: pd.DataFrame) -> str:
    with container_context(
        image="opensearchproject/opensearch:2.11.1",
        ports={9200: 9200, 9600: 9600},
        environment={"discovery.type": "single-node"},
        healthcheck=HealthCheck(
            test="curl --fail https://localhost:9200/_cat/health -ku 'admin:admin' >/dev/null || exit 1",  # noqa: E501
            interval=1,
        ),
    ):
        with get_client() as client:
            Movie.init(using=client)
            for i, row in movies_dataframe.iterrows():
                movie = Movie(
                    meta={"id": i},
                    title=row["Title"],
                    year=row["Release Year"],
                    director=row["Director"],
                    cast=row["Cast"],
                    genre=row["Genre"],
                    wiki_page=row["Wiki Page"],
                    ethnicity=row["Origin/Ethnicity"],
                    plot=row["Plot"],
                )
                movie.save(using=client)
            wait_for_write(
                client=client, index_name=SOURCE_INDEX_NAME, expected_count=len(movies_dataframe)
            )
        yield SOURCE_INDEX_NAME


@pytest.fixture
def destination_index(opensearch_elements_mapping: dict) -> str:
    with container_context(
        image="opensearchproject/opensearch:2.11.1",
        ports={9200: 9200, 9600: 9600},
        environment={"discovery.type": "single-node"},
        healthcheck=HealthCheck(
            test="curl --fail https://localhost:9200/_cat/health -ku 'admin:admin' >/dev/null || exit 1",  # noqa: E501
            interval=1,
        ),
    ):
        with get_client() as client:
            response = client.indices.create(
                index=DESTINATION_INDEX_NAME, body=opensearch_elements_mapping
            )
            if not response["acknowledged"]:
                raise RuntimeError(f"failed to create index: {response}")
        yield DESTINATION_INDEX_NAME


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG)
async def test_opensearch_source(source_index: str, movies_dataframe: pd.DataFrame):
    indexer_config = OpenSearchIndexerConfig(index_name=source_index)
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        connection_config = OpenSearchConnectionConfig(
            access_config=OpenSearchAccessConfig(password="admin"),
            username="admin",
            hosts=["http://localhost:9200"],
            use_ssl=True,
        )
        download_config = OpenSearchDownloaderConfig(download_dir=tempdir_path)
        indexer = OpenSearchIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        downloader = OpenSearchDownloader(
            connection_config=connection_config, download_config=download_config
        )
        expected_num_files = len(movies_dataframe)
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=ValidationConfigs(
                test_id=CONNECTOR_TYPE,
                expected_num_files=expected_num_files,
                expected_number_indexed_file_data=1,
                validate_downloaded_files=True,
            ),
        )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG)
def test_opensearch_source_precheck_fail_no_cluster():
    indexer_config = OpenSearchIndexerConfig(index_name="index")

    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    indexer = OpenSearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG)
def test_opensearch_source_precheck_fail_no_index(source_index: str):
    indexer_config = OpenSearchIndexerConfig(index_name="index")

    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    indexer = OpenSearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
async def test_opensearch_destination(
    upload_file: Path,
    destination_index: str,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    stager = OpenSearchUploadStager(
        upload_stager_config=OpenSearchUploadStagerConfig(index_name=destination_index)
    )

    uploader = OpenSearchUploader(
        connection_config=connection_config,
        upload_config=OpenSearchUploaderConfig(index_name=destination_index),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
def test_opensearch_destination_precheck_fail():
    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    uploader = OpenSearchUploader(
        connection_config=connection_config,
        upload_config=OpenSearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
def test_opensearch_destination_precheck_fail_no_index(destination_index: str):
    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    uploader = OpenSearchUploader(
        connection_config=connection_config,
        upload_config=OpenSearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()



import asyncio
import hashlib
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, TypedDict

from unstructured_ingest.v2.interfaces import FileData
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.pipeline.interfaces import PipelineStep
from unstructured_ingest.v2.processes.embedder import Embedder
from unstructured_ingest.v2.utils import serialize_base_model_json

STEP_ID = "embed"


class EmbedStepResponse(TypedDict):
    file_data_path: str
    path: str


@dataclass
class EmbedStep(PipelineStep):
    process: Embedder
    identifier: str = STEP_ID

    def __str__(self):
        return f"{self.identifier} ({self.process.config.embedding_provider})"

    def __post_init__(self):
        config = self.process.config.model_dump_json() if self.process.config else None
        logger.info(f"created {self.identifier} with configs: {config}")

    def should_embed(self, filepath: Path, file_data: FileData) -> bool:
        if self.context.reprocess or file_data.reprocess:
            return True
        return not filepath.exists()

    def get_output_filepath(self, filename: Path) -> Path:
        hashed_output_file = f"{self.get_hash(extras=[filename.name])}.json"
        filepath = (self.cache_dir / hashed_output_file).resolve()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        return filepath

    def _save_output(self, output_filepath: str, embedded_content: list[dict]):
        with open(str(output_filepath), "w") as f:
            logger.debug(f"writing embedded output to: {output_filepath}")
            json.dump(embedded_content, f, indent=2)

    async def _run_async(self, fn: Callable, path: str, file_data_path: str) -> EmbedStepResponse:
        path = Path(path)
        file_data = FileData.from_file(path=file_data_path)
        output_filepath = self.get_output_filepath(filename=path)
        if not self.should_embed(filepath=output_filepath, file_data=file_data):
            logger.debug(f"skipping embedding, output already exists: {output_filepath}")
            return EmbedStepResponse(file_data_path=file_data_path, path=str(output_filepath))
        fn_kwargs = {"elements_filepath": path}
        if not asyncio.iscoroutinefunction(fn):
            embed_content_raw = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                embed_content_raw = await fn(**fn_kwargs)
        else:
            embed_content_raw = await fn(**fn_kwargs)

        self._save_output(
            output_filepath=str(output_filepath),
            embedded_content=embed_content_raw,
        )
        return EmbedStepResponse(file_data_path=file_data_path, path=str(output_filepath))

    def get_hash(self, extras: Optional[list[str]]) -> str:
        hashable_string = serialize_base_model_json(
            model=self.process.config, sort_keys=True, ensure_ascii=True
        )
        if extras:
            hashable_string += "".join(extras)
        return hashlib.sha256(hashable_string.encode()).hexdigest()[:12]



import tempfile
from contextlib import contextmanager
from pathlib import Path

import pandas as pd
import pytest
from psycopg2 import connect

from test.integration.connectors.utils.constants import DESTINATION_TAG, SOURCE_TAG, env_setup_path
from test.integration.connectors.utils.docker_compose import docker_compose_context
from test.integration.connectors.utils.validation import (
    ValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.v2.interfaces import FileData
from unstructured_ingest.v2.processes.connectors.sql.postgres import (
    CONNECTOR_TYPE,
    PostgresAccessConfig,
    PostgresConnectionConfig,
    PostgresDownloader,
    PostgresDownloaderConfig,
    PostgresIndexer,
    PostgresIndexerConfig,
    PostgresUploader,
    PostgresUploadStager,
)

SEED_DATA_ROWS = 20


@contextmanager
def postgres_download_setup() -> None:
    with docker_compose_context(docker_compose_path=env_setup_path / "sql" / "postgres" / "source"):
        connection = connect(
            user="unstructured",
            password="test",
            dbname="test_db",
            host="localhost",
            port=5433,
        )
        with connection.cursor() as cursor:
            for i in range(SEED_DATA_ROWS):
                sql_statment = f"INSERT INTO cars (brand, price) VALUES " f"('brand_{i}', {i})"
                cursor.execute(sql_statment)
            connection.commit()
        yield


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, "sql")
async def test_postgres_source():
    connect_params = {
        "host": "localhost",
        "port": 5433,
        "database": "test_db",
        "user": "unstructured",
        "password": "test",
    }
    with postgres_download_setup():
        with tempfile.TemporaryDirectory() as tmpdir:
            connection_config = PostgresConnectionConfig(
                host=connect_params["host"],
                port=connect_params["port"],
                database=connect_params["database"],
                username=connect_params["user"],
                access_config=PostgresAccessConfig(password=connect_params["password"]),
            )
            indexer = PostgresIndexer(
                connection_config=connection_config,
                index_config=PostgresIndexerConfig(
                    table_name="cars", id_column="car_id", batch_size=5
                ),
            )
            downloader = PostgresDownloader(
                connection_config=connection_config,
                download_config=PostgresDownloaderConfig(
                    fields=["car_id", "brand"], download_dir=Path(tmpdir)
                ),
            )
            await source_connector_validation(
                indexer=indexer,
                downloader=downloader,
                configs=ValidationConfigs(
                    test_id="postgres",
                    expected_num_files=SEED_DATA_ROWS,
                    expected_number_indexed_file_data=4,
                    validate_downloaded_files=True,
                ),
            )


def validate_destination(
    connect_params: dict,
    expected_num_elements: int,
    test_embedding: list[float],
    expected_text: str,
):
    # Run the following validations:
    # * Check that the number of records in the table match the expected value
    # * Given the embedding, make sure it matches the associated text it belongs to
    with connect(**connect_params) as connection:
        cursor = connection.cursor()
        query = "select count(*) from elements;"
        cursor.execute(query)
        count = cursor.fetchone()[0]
        assert (
            count == expected_num_elements
        ), f"dest check failed: got {count}, expected {expected_num_elements}"

        cursor.execute("SELECT embeddings FROM elements order by text limit 1")
        similarity_query = (
            f"SELECT text FROM elements ORDER BY embeddings <-> '{test_embedding}' LIMIT 1;"
        )
        cursor.execute(similarity_query)
        res = cursor.fetchone()
        assert res[0] == expected_text


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, "sql")
async def test_postgres_destination(upload_file: Path):
    # the postgres destination connector doesn't leverage the file data but is required as an input,
    # mocking it with arbitrary values to meet the base requirements:
    mock_file_data = FileData(identifier="mock file data", connector_type=CONNECTOR_TYPE)
    with docker_compose_context(
        docker_compose_path=env_setup_path / "sql" / "postgres" / "destination"
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            stager = PostgresUploadStager()
            stager_params = {
                "elements_filepath": upload_file,
                "file_data": mock_file_data,
                "output_dir": Path(tmpdir),
                "output_filename": "test_db",
            }
            if stager.is_async():
                staged_path = await stager.run_async(**stager_params)
            else:
                staged_path = stager.run(**stager_params)

            # The stager should append the `.json` suffix to the output filename passed in.
            assert staged_path.name == "test_db.json"

            connect_params = {
                "host": "localhost",
                "port": 5433,
                "database": "elements",
                "user": "unstructured",
                "password": "test",
            }

            uploader = PostgresUploader(
                connection_config=PostgresConnectionConfig(
                    host=connect_params["host"],
                    port=connect_params["port"],
                    database=connect_params["database"],
                    username=connect_params["user"],
                    access_config=PostgresAccessConfig(password=connect_params["password"]),
                )
            )

            uploader.run(path=staged_path, file_data=mock_file_data)

            staged_df = pd.read_json(staged_path, orient="records", lines=True)
            sample_element = staged_df.iloc[0]
            expected_num_elements = len(staged_df)
            validate_destination(
                connect_params=connect_params,
                expected_num_elements=expected_num_elements,
                expected_text=sample_element["text"],
                test_embedding=sample_element["embeddings"],
            )

            uploader.run(path=staged_path, file_data=mock_file_data)
            validate_destination(
                connect_params=connect_params,
                expected_num_elements=expected_num_elements,
                expected_text=sample_element["text"],
                test_embedding=sample_element["embeddings"],
            )



# ruff: noqa: I001
import json
import tempfile
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

import pandas as pd
import pytest
from elasticsearch import Elasticsearch as ElasticsearchClient
from elasticsearch.helpers import bulk

from test.integration.connectors.utils.constants import DESTINATION_TAG, SOURCE_TAG
from test.integration.connectors.utils.docker import HealthCheck, container_context
from test.integration.connectors.utils.validation import (
    ValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.elasticsearch.elasticsearch import (
    CONNECTOR_TYPE,
    ElasticsearchAccessConfig,
    ElasticsearchConnectionConfig,
    ElasticsearchDownloader,
    ElasticsearchDownloaderConfig,
    ElasticsearchIndexer,
    ElasticsearchIndexerConfig,
    ElasticsearchUploader,
    ElasticsearchUploaderConfig,
    ElasticsearchUploadStager,
    ElasticsearchUploadStagerConfig,
)

SOURCE_INDEX_NAME = "movies"
DESTINATION_INDEX_NAME = "elements"
ES_USERNAME = "elastic"
ES_PASSWORD = "elastic_password"


@contextmanager
def get_client() -> Generator[ElasticsearchClient, None, None]:
    with ElasticsearchClient(
        hosts="http://localhost:9200", basic_auth=(ES_USERNAME, ES_PASSWORD), request_timeout=30
    ) as client:
        yield client


def form_elasticsearch_doc_dict(i, csv_row):
    return {
        "_index": SOURCE_INDEX_NAME,
        "_id": i,
        "_source": {
            "title": csv_row["Title"],
            "ethnicity": csv_row["Origin/Ethnicity"],
            "director": csv_row["Director"],
            "cast": csv_row["Cast"],
            "genre": csv_row["Genre"],
            "plot": csv_row["Plot"],
            "year": csv_row["Release Year"],
            "wiki_page": csv_row["Wiki Page"],
        },
    }


def dataframe_to_upload_data(df: pd.DataFrame) -> list[dict]:
    upload_data = []
    for index, row in df.iterrows():
        upload_data.append(form_elasticsearch_doc_dict(index, row))
    return upload_data


def get_index_count(client: ElasticsearchClient, index_name: str) -> int:
    count_resp = client.cat.count(index=index_name, format="json")
    return int(count_resp[0]["count"])


def validate_count(
    client: ElasticsearchClient,
    index_name: str,
    expected_count: int,
    retries: int = 10,
    interval: int = 1,
) -> None:
    current_count = get_index_count(client, index_name)
    if current_count == expected_count:
        return
    tries = 0
    while tries < retries:
        print(
            f"retrying validation to check if expected count "
            f"{expected_count} will match current count {current_count}"
        )
        time.sleep(interval)
        current_count = get_index_count(client, index_name)
        if current_count == expected_count:
            break
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from index: {current_count}"
    )


def seed_source_db(df: pd.DataFrame):
    mapping = {
        "properties": {
            "title": {"type": "text", "analyzer": "english"},
            "ethnicity": {"type": "text", "analyzer": "standard"},
            "director": {"type": "text", "analyzer": "standard"},
            "cast": {"type": "text", "analyzer": "standard"},
            "genre": {"type": "text", "analyzer": "standard"},
            "plot": {"type": "text", "analyzer": "english"},
            "year": {"type": "integer"},
            "wiki_page": {"type": "keyword"},
        },
    }
    # seed content
    with get_client() as client:
        client.indices.create(index=SOURCE_INDEX_NAME, mappings=mapping)
        upload_data = dataframe_to_upload_data(df=df)
        bulk(client, upload_data)
        client.indices.refresh(index=SOURCE_INDEX_NAME)
        count = get_index_count(client, SOURCE_INDEX_NAME)
        print(f"seeded {SOURCE_INDEX_NAME} index with {count} records")


@pytest.fixture
def source_index(movies_dataframe: pd.DataFrame) -> str:
    with container_context(
        image="docker.elastic.co/elasticsearch/elasticsearch:8.7.0",
        ports={9200: 9200, 9300: 9300},
        environment={
            "discovery.type": "single-node",
            "xpack.security.enabled": True,
            "ELASTIC_PASSWORD": ES_PASSWORD,
            "ELASTIC_USER": ES_USERNAME,
        },
        healthcheck=HealthCheck(
            test="curl --silent --fail -u ${ELASTIC_USER}:${ELASTIC_PASSWORD} localhost:9200/_cluster/health || exit 1",  # noqa: E501
            interval=1,
            start_period=5,
        ),
    ):
        seed_source_db(df=movies_dataframe)
        yield SOURCE_INDEX_NAME


@pytest.fixture
def destination_index(elasticsearch_elements_mapping: dict) -> str:
    with container_context(
        image="docker.elastic.co/elasticsearch/elasticsearch:8.7.0",
        ports={9200: 9200, 9300: 9300},
        environment={
            "discovery.type": "single-node",
            "xpack.security.enabled": True,
            "ELASTIC_PASSWORD": ES_PASSWORD,
            "ELASTIC_USER": ES_USERNAME,
        },
        healthcheck=HealthCheck(
            test="curl --silent --fail -u ${ELASTIC_USER}:${ELASTIC_PASSWORD} localhost:9200/_cluster/health || exit 1",  # noqa: E501
            interval=1,
            start_period=5,
        ),
    ):
        with get_client() as client:
            response = client.indices.create(
                index=DESTINATION_INDEX_NAME, mappings=elasticsearch_elements_mapping
            )
            if not response["acknowledged"]:
                raise RuntimeError(f"failed to create index: {response}")
        yield DESTINATION_INDEX_NAME


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG)
async def test_elasticsearch_source(source_index: str, movies_dataframe: pd.DataFrame):
    indexer_config = ElasticsearchIndexerConfig(index_name=source_index)
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        connection_config = ElasticsearchConnectionConfig(
            access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
            username=ES_USERNAME,
            hosts=["http://localhost:9200"],
        )
        download_config = ElasticsearchDownloaderConfig(download_dir=tempdir_path)
        indexer = ElasticsearchIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        downloader = ElasticsearchDownloader(
            connection_config=connection_config, download_config=download_config
        )
        expected_num_files = len(movies_dataframe)
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=ValidationConfigs(
                test_id=CONNECTOR_TYPE,
                expected_num_files=expected_num_files,
                expected_number_indexed_file_data=1,
                validate_downloaded_files=True,
            ),
        )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG)
def test_elasticsearch_source_precheck_fail_no_cluster():
    indexer_config = ElasticsearchIndexerConfig(index_name="index")

    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    indexer = ElasticsearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG)
def test_elasticsearch_source_precheck_fail_no_index(source_index: str):
    indexer_config = ElasticsearchIndexerConfig(index_name="index")

    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    indexer = ElasticsearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
async def test_elasticsearch_destination(
    upload_file: Path,
    destination_index: str,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    stager = ElasticsearchUploadStager(
        upload_stager_config=ElasticsearchUploadStagerConfig(index_name=destination_index)
    )

    uploader = ElasticsearchUploader(
        connection_config=connection_config,
        upload_config=ElasticsearchUploaderConfig(index_name=destination_index),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
def test_elasticsearch_destination_precheck_fail():
    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    uploader = ElasticsearchUploader(
        connection_config=connection_config,
        upload_config=ElasticsearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
def test_elasticsearch_destination_precheck_fail_no_index(destination_index: str):
    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    uploader = ElasticsearchUploader(
        connection_config=connection_config,
        upload_config=ElasticsearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()



from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field

from unstructured_ingest.embed.openai import OpenAIEmbeddingConfig, OpenAIEmbeddingEncoder
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from openai import AzureOpenAI


class AzureOpenAIEmbeddingConfig(OpenAIEmbeddingConfig):
    api_version: str = Field(description="Azure API version", default="2024-06-01")
    azure_endpoint: str
    embedder_model_name: str = Field(default="text-embedding-ada-002", alias="model_name")

    @requires_dependencies(["openai"], extras="openai")
    def get_client(self) -> "AzureOpenAI":
        from openai import AzureOpenAI

        return AzureOpenAI(
            api_key=self.api_key.get_secret_value(),
            api_version=self.api_version,
            azure_endpoint=self.azure_endpoint,
        )


@dataclass
class AzureOpenAIEmbeddingEncoder(OpenAIEmbeddingEncoder):
    config: AzureOpenAIEmbeddingConfig



import json
import math
import os
import re
import time
from pathlib import Path
from typing import Generator
from uuid import uuid4

import pytest
from pinecone import Pinecone, ServerlessSpec
from pinecone.core.openapi.shared.exceptions import NotFoundException

from test.integration.connectors.utils.constants import (
    DESTINATION_TAG,
)
from test.integration.utils import requires_env
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connectors.pinecone import (
    CONNECTOR_TYPE,
    MAX_QUERY_RESULTS,
    PineconeAccessConfig,
    PineconeConnectionConfig,
    PineconeUploader,
    PineconeUploaderConfig,
    PineconeUploadStager,
    PineconeUploadStagerConfig,
)

METADATA_BYTES_LIMIT = (
    40960  # 40KB https://docs.pinecone.io/reference/quotas-and-limits#hard-limits
)
VECTOR_DIMENSION = 384
SPEC = {"serverless": {"cloud": "aws", "region": "us-east-1"}}
ALLOWED_METADATA_FIELD = "text"
API_KEY = "PINECONE_API_KEY"


def get_api_key() -> str:
    api_key = os.getenv(API_KEY, None)
    assert api_key
    return api_key


def wait_for_delete(client: Pinecone, index_name: str, timeout=60, interval=1) -> None:
    start = time.time()
    while True and time.time() - start < timeout:
        try:
            description = client.describe_index(name=index_name)
            logger.info(f"current index status: {description}")
        except NotFoundException:
            return
        time.sleep(interval)

    raise TimeoutError("time out waiting for index to delete")


def wait_for_ready(client: Pinecone, index_name: str, timeout=60, interval=1) -> None:
    def is_ready_status():
        description = client.describe_index(name=index_name)
        status = description["status"]
        return status["ready"]

    start = time.time()
    is_ready = is_ready_status()
    while not is_ready and time.time() - start < timeout:
        time.sleep(interval)
        is_ready = is_ready_status()
    if not is_ready:
        raise TimeoutError("time out waiting for index to be ready")


@pytest.fixture
def pinecone_index() -> Generator[str, None, None]:
    pinecone = Pinecone(api_key=get_api_key())
    random_id = str(uuid4()).split("-")[0]
    index_name = f"ingest-test-{random_id}"
    assert len(index_name) < 45
    logger.info(f"Creating index: {index_name}")
    try:
        pinecone.create_index(
            name=index_name,
            dimension=384,
            metric="cosine",
            spec=ServerlessSpec(
                cloud="aws",
                region="us-east-1",
            ),
            deletion_protection="disabled",
        )
        wait_for_ready(client=pinecone, index_name=index_name)
        yield index_name
    except Exception as e:
        logger.error(f"failed to create index {index_name}: {e}")
    finally:
        try:
            logger.info(f"deleting index: {index_name}")
            pinecone.delete_index(name=index_name)
            wait_for_delete(client=pinecone, index_name=index_name)
        except NotFoundException:
            return


def validate_pinecone_index(
    index_name: str, expected_num_of_vectors: int, retries=30, interval=1
) -> None:
    # Because there's a delay for the index to catch up to the recent writes, add in a retry
    pinecone = Pinecone(api_key=get_api_key())
    index = pinecone.Index(name=index_name)
    vector_count = -1
    for i in range(retries):
        index_stats = index.describe_index_stats()
        vector_count = index_stats["total_vector_count"]
        if vector_count == expected_num_of_vectors:
            logger.info(f"expected {expected_num_of_vectors} == vector count {vector_count}")
            break
        logger.info(
            f"retry attempt {i}: expected {expected_num_of_vectors} != vector count {vector_count}"
        )
        time.sleep(interval)
    assert vector_count == expected_num_of_vectors, (
        f"vector count from index ({vector_count}) doesn't "
        f"match expected number: {expected_num_of_vectors}"
    )


@requires_env(API_KEY)
@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
async def test_pinecone_destination(pinecone_index: str, upload_file: Path, temp_dir: Path):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="pinecone_mock_id",
    )
    connection_config = PineconeConnectionConfig(
        index_name=pinecone_index,
        access_config=PineconeAccessConfig(api_key=get_api_key()),
    )
    stager_config = PineconeUploadStagerConfig()
    stager = PineconeUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=upload_file,
        output_dir=temp_dir,
        output_filename=upload_file.name,
        file_data=file_data,
    )

    upload_config = PineconeUploaderConfig()
    uploader = PineconeUploader(connection_config=connection_config, upload_config=upload_config)
    uploader.precheck()

    uploader.run(path=new_upload_file, file_data=file_data)
    with new_upload_file.open() as f:
        staged_content = json.load(f)
    expected_num_of_vectors = len(staged_content)
    logger.info("validating first upload")
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )

    # Rerun uploader and make sure no duplicates exist
    uploader.run(path=new_upload_file, file_data=file_data)
    logger.info("validating second upload")
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )


@requires_env(API_KEY)
@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
@pytest.mark.skip(reason="TODO: get this to work")
async def test_pinecone_destination_large_index(
    pinecone_index: str, upload_file: Path, temp_dir: Path
):
    new_file = temp_dir / "large_file.json"
    with upload_file.open() as f:
        upload_content = json.load(f)

    min_entries = math.ceil((MAX_QUERY_RESULTS * 2) / len(upload_content))
    new_content = (upload_content * min_entries)[: (2 * MAX_QUERY_RESULTS)]
    print(f"Creating large index content with {len(new_content)} records")
    with new_file.open("w") as f:
        json.dump(new_content, f)

    expected_num_of_vectors = len(new_content)
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=new_file.name, filename=new_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="pinecone_mock_id",
    )
    connection_config = PineconeConnectionConfig(
        index_name=pinecone_index,
        access_config=PineconeAccessConfig(api_key=get_api_key()),
    )
    stager_config = PineconeUploadStagerConfig()
    stager = PineconeUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=new_file,
        output_dir=temp_dir,
        output_filename=new_file.name,
        file_data=file_data,
    )

    upload_config = PineconeUploaderConfig()
    uploader = PineconeUploader(connection_config=connection_config, upload_config=upload_config)
    uploader.precheck()

    uploader.run(path=new_upload_file, file_data=file_data)
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )
    # Rerun uploader and make sure no duplicates exist
    uploader.run(path=new_upload_file, file_data=file_data)
    logger.info("validating second upload")
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )


@requires_env(API_KEY)
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG)
def test_large_metadata(pinecone_index: str, tmp_path: Path, upload_file: Path):
    stager = PineconeUploadStager()
    uploader = PineconeUploader(
        connection_config=PineconeConnectionConfig(
            access_config=PineconeAccessConfig(api_key=get_api_key()),
            index_name=pinecone_index,
        ),
        upload_config=PineconeUploaderConfig(),
    )
    large_metadata_upload_file = tmp_path / "mock-upload-file.pdf.json"
    large_metadata = {ALLOWED_METADATA_FIELD: "0" * 2 * METADATA_BYTES_LIMIT}

    with open(upload_file) as file:
        elements = json.load(file)

    with open(large_metadata_upload_file, "w") as file:
        mock_element = elements[0]
        mock_element["metadata"] = large_metadata
        json.dump([mock_element], file)

    file_data = FileData(
        source_identifiers=SourceIdentifiers(
            fullpath=large_metadata_upload_file.name, filename=large_metadata_upload_file.name
        ),
        connector_type=CONNECTOR_TYPE,
        identifier="mock-file-data",
    )
    staged_file = stager.run(
        file_data, large_metadata_upload_file, tmp_path, large_metadata_upload_file.name
    )
    try:
        uploader.run(staged_file, file_data)
    except DestinationConnectionError as e:
        error_line = r"Metadata size is \d+ bytes, which exceeds the limit of \d+ bytes per vector"
        if re.search(re.compile(error_line), str(e)) is None:
            raise e
        raise pytest.fail("Upload request failed due to metadata exceeding limits.")

    validate_pinecone_index(pinecone_index, 1, interval=5)
